"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const bodyParser = require("body-parser");
const express_1 = require("express");
const ooth_i18n_1 = require("ooth-i18n");
const passport = require("passport");
const expressSession = require("express-session");
const cookieParser = require("cookie-parser");
const { Strategy } = require('passport-custom');
const locale = require('locale');
const DEFAULT_LANGUAGE = 'en';
const DEFAULT_TRANSLATIONS = {
    en: require('../i18n/en.json'),
};
const authenticate = (methodName, session, req, res) => new Promise((resolve, reject) => passport.authenticate(methodName, {
    session,
}, (err, user, info) => err ? reject(err) : !user && info && info.message ? reject(info.message) : resolve(user))(req, res));
class Ooth {
    constructor({ app, backend, path, defaultLanguage, translations, session, sessionSecret }) {
        this.uniqueFields = {};
        this.strategies = {};
        this.route = express_1.Router();
        this.middleware = [];
        this.afterware = [];
        this.authAfterware = [];
        this.listeners = {};
        this.getUniqueField = (user, fieldName) => {
            if (this.uniqueFields[fieldName]) {
                for (const strategyName of Object.keys(this.uniqueFields[fieldName])) {
                    const actualFieldName = this.uniqueFields[fieldName][strategyName];
                    if (user[strategyName] && user[strategyName][actualFieldName]) {
                        return user[strategyName][actualFieldName];
                    }
                }
            }
            return null;
        };
        this.getUserByFields = (strategyName, fields) => __awaiter(this, void 0, void 0, function* () {
            const actualFields = {};
            Object.keys(fields).forEach((field) => {
                actualFields[`${strategyName}.${field}`] = fields[field];
            });
            return yield this.backend.getUser(actualFields);
        });
        this.requireLogged = (req, res, next) => {
            if (!req.user) {
                return res.status(400).send({
                    status: 'error',
                    message: this.__('not_logged', null, req.locale),
                });
            }
            next();
        };
        this.requireNotLogged = (req, res, next) => {
            if (req.user) {
                return res.status(400).send({
                    status: 'error',
                    message: this.__('already_logged', null, req.locale),
                });
            }
            next();
        };
        this.requireNotRegistered = (req, res, next) => {
            if (req.user) {
                return res.status(400).send({
                    status: 'error',
                    message: this.__('already_registered', null, req.locale),
                });
            }
            next();
        };
        this.requireRegisteredWith = (strategy) => {
            return (req, res, next) => this.requireLogged(req, res, () => __awaiter(this, void 0, void 0, function* () {
                const user = yield this.getUserById(req.user);
                if (!user[strategy]) {
                    return res.status(400).send({
                        status: 'error',
                        message: this.__('not_registered_with', { strategy }, req.locale),
                    });
                }
                next();
            }));
        };
        this.getUserById = (id) => __awaiter(this, void 0, void 0, function* () {
            return this.backend.getUserById(id);
        });
        this.getUserByUniqueField = (fieldName, value) => __awaiter(this, void 0, void 0, function* () {
            return yield this.backend.getUserByValue(Object.keys(this.uniqueFields[fieldName]).map((strategyName) => `${strategyName}.${this.uniqueFields[fieldName][strategyName]}`), value);
        });
        this.updateUser = (strategyName, id, fields) => __awaiter(this, void 0, void 0, function* () {
            const actualFields = {};
            Object.keys(fields).forEach((field) => {
                actualFields[`${strategyName}.${field}`] = fields[field];
            });
            return yield this.backend.updateUser(id, actualFields);
        });
        this.insertUser = (strategyName, fields) => __awaiter(this, void 0, void 0, function* () {
            const query = {};
            if (fields) {
                query[strategyName] = fields;
            }
            return yield this.backend.insertUser(query);
        });
        if (!app) {
            throw new Error('App is required.');
        }
        this.app = app;
        if (!backend) {
            throw new Error('Backend is required.');
        }
        this.backend = backend;
        this.path = path || '/';
        this.translations = translations || DEFAULT_TRANSLATIONS;
        this.defaultLanguage = defaultLanguage || DEFAULT_LANGUAGE;
        this.__ = ooth_i18n_1.getI18n(this.translations, this.defaultLanguage);
        if (sessionSecret && session) {
            throw new Error('Either set session or sessionSecret, not both');
        }
        if (session) {
            this.session = session;
        }
        if (sessionSecret) {
            this.session = expressSession({
                secret: sessionSecret,
                name: 'ooth-session-id',
                resave: false,
                saveUninitialized: true,
            });
        }
        if (this.session) {
            this.registerAuthAfterware((result, _userId, req) => __awaiter(this, void 0, void 0, function* () {
                if (req.body.remember) {
                    req.session.cookie.maxAge = 30 * 24 * 60 * 60 * 1000;
                }
                return result;
            }));
        }
        // App-wide configuration
        this.app.use(bodyParser.json());
        this.app.use(locale(Object.keys(this.translations), this.defaultLanguage));
        if (this.session) {
            this.app.use(cookieParser());
            this.app.use(this.session);
        }
        this.app.use(passport.initialize());
        if (this.session) {
            this.app.use(passport.session());
            passport.serializeUser((id, done) => done(null, id));
            passport.deserializeUser((id, done) => done(null, id));
            this.registerPrimaryAuth('session', 'logout', [this.requireLogged], new Strategy((_req, done) => done(null, null)));
        }
        this.app.use((req, res, next) => __awaiter(this, void 0, void 0, function* () {
            try {
                const fullRequest = req;
                for (const strategyName of Object.keys(this.strategies)) {
                    for (const methodName of Object.keys(this.strategies[strategyName].secondaryAuth)) {
                        if (this.strategies[strategyName].secondaryAuth[methodName](fullRequest)) {
                            const userId = yield authenticate(`${strategyName}-${methodName}`, !!this.session, fullRequest, res);
                            yield new Promise((res, rej) => fullRequest.login(userId, { session: !!this.session }, (e) => (e ? rej(e) : res())));
                        }
                    }
                }
                next();
            }
            catch (e) {
                console.error(e);
                return res.status(400).send({
                    status: 'error',
                    message: typeof e === 'string' ? e : e.message,
                });
            }
        }));
        this.app.use(this.path, this.route);
    }
    usesSession() {
        return !!this.session;
    }
    registerUniqueField(strategyName, id, fieldName) {
        if (!this.uniqueFields[id]) {
            this.uniqueFields[id] = {};
        }
        this.uniqueFields[id][strategyName] = fieldName;
    }
    registerStrategyUniqueField(strategyName, ...fieldNames) {
        for (const fieldName of fieldNames) {
            this.getStrategy(strategyName).uniqueFields[fieldName] = true;
        }
    }
    registerProfileFields(strategyName, ...fieldNames) {
        for (const fieldName of fieldNames) {
            this.getStrategy(strategyName).profileFields[fieldName] = true;
        }
    }
    registerPrimaryAuth(strategyName, method, middleware, strategy, resolveUserId = (_req, userId) => __awaiter(this, void 0, void 0, function* () { return userId; })) {
        const methodName = `${strategyName}-${method}`;
        const routeName = `/${strategyName}/${method}`;
        passport.use(methodName, strategy);
        this.route.post(routeName, ...middleware, (req, res) => __awaiter(this, void 0, void 0, function* () {
            const fullRequest = req;
            try {
                const authResult = yield authenticate(methodName, !!this.session, fullRequest, res);
                const userId = yield resolveUserId(fullRequest, authResult);
                if (req.user !== userId) {
                    if (userId) {
                        this.emit('ooth', 'login', { userId, sessionId: req.session && req.session.id });
                        yield new Promise((res, rej) => fullRequest.login(userId, { session: !!this.session }, (e) => (e ? rej(e) : res())));
                    }
                    else {
                        this.emit('ooth', 'logout', { userId: req.user, sessionId: req.session && req.session.id });
                        req.logout();
                    }
                }
                let result = {};
                for (const aw of [...this.authAfterware, ...this.afterware]) {
                    result = yield aw(result, fullRequest.user, fullRequest, res);
                }
                return res.send(result);
            }
            catch (e) {
                console.error(e);
                return res.status(400).send({
                    status: 'error',
                    message: e.message || e,
                });
            }
        }));
    }
    registerPrimaryConnect(strategyName, method, middleware, strategy) {
        this.registerPrimaryAuth(strategyName, method, middleware, strategy, (req, userPart) => __awaiter(this, void 0, void 0, function* () {
            if (!userPart) {
                return;
            }
            let userId;
            for (const field of Object.keys(this.getStrategy(strategyName).uniqueFields)) {
                const value = userPart[field];
                if (value) {
                    const userCandidate = yield this.backend.getUser({
                        [`${strategyName}.${field}`]: value,
                    });
                    if (userCandidate) {
                        if (!userId || userId === userCandidate._id) {
                            userId = userCandidate._id;
                        }
                        else {
                            throw new Error(this.__('register.ambiguous_authentication_strategy', null, req.locale));
                        }
                    }
                }
            }
            for (const field of Object.keys(this.uniqueFields)) {
                if (this.uniqueFields[field][strategyName]) {
                    const value = userPart[this.uniqueFields[field][strategyName]];
                    if (value) {
                        const userCandidate = yield this.getUserByUniqueField(field, value);
                        if (userCandidate) {
                            if (!userId || userId === userCandidate._id) {
                                userId = userCandidate._id;
                            }
                            else {
                                throw new Error(this.__('register.ambiguous_authentication_strategy', null, req.locale));
                            }
                        }
                    }
                }
            }
            let registered = false;
            if (req.user) {
                // User is already logged in
                if (userId && userId !== req.user) {
                    throw new Error(this.__('register.foreign_authentication_strategy', null, req.locale));
                }
                userId = req.user;
                // Update user
                yield this.updateUser(strategyName, userId, userPart);
            }
            else {
                // User hasn't logged in
                if (userId) {
                    // User exists already, update
                    yield this.updateUser(strategyName, userId, userPart);
                }
                else {
                    // Need to create a new user
                    userId = yield this.insertUser(strategyName, userPart);
                    registered = true;
                }
            }
            if (registered) {
                this.emit('ooth', 'register', { userId });
            }
            return userId;
        }));
    }
    registerSecondaryAuth(strategyName, method, condition, strategy) {
        const methodName = `${strategyName}-${method}`;
        if (this.getStrategy(strategyName).secondaryAuth[method]) {
            throw new Error(`Secondary auth ${methodName} has already been registered.`);
        }
        passport.use(methodName, strategy);
        this.getStrategy(strategyName).secondaryAuth[method] = condition;
    }
    registerMethod(strategyName, method, middleware, handler, httpMethod = 'POST') {
        this.getStrategy(strategyName).methods[method] = handler;
        const finalHandler = (req, res) => __awaiter(this, void 0, void 0, function* () {
            try {
                let result = yield this.callMethod(strategyName, method, req.body, req.user, req.locale);
                for (const aw of this.afterware) {
                    result = yield aw(result, req.user, req, res);
                }
                res.send(result);
            }
            catch (e) {
                console.error(e);
                return res.status(400).send({
                    status: 'error',
                    message: e.message,
                });
            }
        });
        this.route[httpMethod === 'GET' ? 'get' : 'post'](`/${strategyName}/${method}`, ...middleware, finalHandler);
    }
    callMethod(strategyName, method, params, user, locale) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.getStrategy(strategyName).methods[method](params, user, locale);
        });
    }
    registerMiddleware(...middleware) {
        this.middleware.push(...middleware);
    }
    registerAfterware(...afterware) {
        this.afterware.push(...afterware);
    }
    registerAuthAfterware(...afterware) {
        this.authAfterware.push(...afterware);
    }
    getProfile(user) {
        if (!user) {
            return;
        }
        const profile = {
            _id: user._id,
        };
        for (const strategyName of Object.keys(this.strategies)) {
            if (!user[strategyName]) {
                continue;
            }
            for (const fieldName of Object.keys(this.getStrategy(strategyName).profileFields)) {
                if (!profile[strategyName]) {
                    profile[strategyName] = {};
                }
                profile[strategyName][fieldName] = user[strategyName][fieldName];
            }
        }
        return profile;
    }
    getApp() {
        return this.app;
    }
    getRoute() {
        return this.route;
    }
    on(strategyName, eventName, eventListener) {
        const fullEventName = `${strategyName}--${eventName}`;
        if (!this.listeners[fullEventName]) {
            this.listeners[fullEventName] = [];
        }
        this.listeners[fullEventName].push(eventListener);
    }
    emit(strategyName, eventName, payload) {
        return __awaiter(this, void 0, void 0, function* () {
            const fullEventName = `${strategyName}--${eventName}`;
            if (this.listeners[fullEventName]) {
                for (const listener of this.listeners[fullEventName]) {
                    yield listener(payload);
                }
            }
        });
    }
    getStrategy(strategyName) {
        if (!this.strategies[strategyName]) {
            this.strategies[strategyName] = {
                methods: {},
                profileFields: {},
                uniqueFields: {},
                secondaryAuth: {},
            };
        }
        return this.strategies[strategyName];
    }
}
exports.Ooth = Ooth;
